# -*- coding: utf-8 -*-
"""
app.tasks
~~~~~~~~~

This module implements tasks to run.

:copyright: (c) 2018 European Spallation Source ERIC
:license: BSD 2-Clause, see LICENSE for more details.

"""
import tower_cli
from flask import current_app
from rq import Queue


def trigger_vm_creation(name, interface, memory, cores):
    """Trigger a job to create a virtual machine or virtual IOC"""
    extra_vars = [
        f'vmname={name}',
        f'memory={memory}',
        f'cores={cores}',
        f'vcpus={cores}',
        f'vlan_name={interface.network.vlan_name}',
        f'vlan_id={interface.network.vlan_id}',
        f'mac={interface.mac.address}',
    ]
    if interface.is_ioc:
        job_template = current_app.config['AWX_CREATE_VIOC']
    else:
        job_template = current_app.config['AWX_CREATE_VM']
        extra_vars.extend([
            f'ip_address={interface.ip}',
            f'domain={interface.network.domain.name}',
            f'dns={current_app.config["VM_DEFAULT_DNS"]}',
            f'netmask={interface.network.netmask}',
            f'gateway={interface.network.gateway}',
        ])
    q = Queue()
    current_app.logger.info(f'Launch new job to create the {name} VM: {job_template} with {extra_vars}')
    job = q.enqueue(
        launch_job_template,
        job_template=job_template,
        extra_vars=extra_vars,
        timeout=500,
    )
    return job


def trigger_core_services_update():
    """Trigger a job to update the core services on the TN (DNS and DHCP)

    This function should be called every time an host or interface is created/edited

    We can have one running job + one in queue to apply the latest changes.
    Make sure that we don't have more than one in queue.
    """
    job_template = current_app.config['AWX_CORE_SERVICES_UPDATE']
    q = Queue()
    # Only trigger a new job if none is already waiting in queue
    for job in q.jobs:
        if (job.func_name == 'app.tasks.launch_job_template' and
                job.kwargs.get('job_template', '') == job_template):
            current_app.logger.info(f'Already one {job_template} job in queue. No need to trigger a new one.')
            return None
    current_app.logger.info(f'Launch new job to update core services: {job_template}')
    job = q.enqueue(
        launch_job_template,
        job_template=job_template,
    )
    return job


def launch_job_template(job_template, monitor=True, **kwargs):
    if (job_template in (current_app.config['AWX_CREATE_VIOC'], current_app.config['AWX_CREATE_VM']) and
            not current_app.config.get('AWX_VM_CREATION_ENABLED', False)):
        current_app.logger.info('AWX VM creation is disabled. Not sending any request.')
        return 'AWX VM creation not triggered'
    if not current_app.config.get('AWX_JOB_ENABLED', False):
        current_app.logger.info('AWX job is disabled. Not sending any request.')
        return 'AWX job not triggered'
    resource = tower_cli.get_resource('job')
    result = resource.launch(job_template=job_template, monitor=monitor, **kwargs)
    return result
